#!/usr/bin/env python
#
# This file is part of the Score-P software (http://www.score-p.org)
#
# Copyright (c) 2015-2016, 2021,
# Technische Universitaet Dresden, Germany
#
# This software may be modified and distributed under the terms of
# a BSD-style license.  See the COPYING file in the package base
# directory for details.
#

import argparse
import itertools
import os

import _otf2

# ___ Global variables. ____________________________________________________

# @internal
#  @brief Defines the number of locations to generate. */
otf2_LOCATIONS = 4

# @internal
#  @brief Defines the number of events to generate per location. */
otf2_EVENTS = 1

# @internal
#  @brief Defines if to generate MPI-only traces. This default if nothing is
# specified in command line parameters. */
otf2_MPI = True

# @internal
#  @brief Defines if to generate OpenMP-only traces. */
otf2_OPENMP = False

# @internal
#  @brief Defines if to generate hybrid (MPI and OpenMP) traces. */
otf2_HYBRID = False

# @internal
#  @brief Defines if to generate additional attributes to events. */
otf2_ATTRIBUTES = False


def get_time():
    get_time.counter += 1
    return get_time.counter


get_time.counter = -1


def pre_flush(user_data, file_type, location, caller_data, final):
    del user_data, file_type, location, caller_data, final
    return _otf2.FLUSH


def post_flush(user_data, file_type, location):
    del user_data, file_type, location
    return get_time()


def gpot(v):
    """ Calculate the greates power-of-two number which is less/equal to @a v. """
    p = v - 1
    p |= p >> 1
    p |= p >> 2
    p |= p >> 4
    p |= p >> 8
    p |= p >> 16
    p += 1
    if v < p:
        p >>= 1
    return p


# empty string definition */
STRING_EMPTY = 0

# definition IDs for regions */
REGION_MAIN = 0
REGION_OPENMP_PARALLEL = 1
REGION_OPENMP_LOOP = 2
REGION_OPENMP_IMPLICIT_BARRIER = 3
REGION_MPI_SEND = 4
REGION_MPI_RECV = 5
REGION_MPI_BARRIER = 6

# definition IDs for Groups */
GROUP_MPI_LOCATIONS = 0
GROUP_MPI_COMM_WORLD = 1
GROUP_MPI_COMM_SELF = 2
GROUP_ALL_LOCATIONS = 3
GROUP_MPI_COMM_CUBE = 4

# definition IDs for MPI comms */
MPI_COMM_MPI_COMM_WORLD = 0
MPI_COMM_MPI_COMM_SELF = 1
MPI_COMM_MPI_COMM_CUBE = 2

# definition IDs for metrics */
METRIC_A = 0
METRIC_B = 1
METRIC_C = 2
METRIC_D = 3
METRIC_E = 4
NUM_OF_METRICS = 5

# definition IDs for metric classes and instances */
METRIC_CLASS_1 = 0
METRIC_CLASS_2 = 1
METRIC_CLASS_3 = 2
METRIC_CLASS_4 = 3
METRIC_INSTANCE_1 = 4
NUM_OF_CLASSES = 5


def next_value_attr_1():
    next_value_attr_1.value.uint64 += 1
    return next_value_attr_1.value


next_value_attr_1.value = _otf2.AttributeValue()
next_value_attr_1.value.uint64 = 12345


def next_value_attr_2():
    next_value_attr_2.value.int16 += 1
    return next_value_attr_2.value


next_value_attr_2.value = _otf2.AttributeValue()
next_value_attr_2.value.int16 = 23456


# attributes */
class AttributeDef:
    def __init__(self, name, otf2_type, next_value):
        self.name = name
        self.type = otf2_type
        self.next_value = next_value


attribute_defs = [AttributeDef(name="attribute 1",
                               otf2_type=_otf2.TYPE_UINT64,
                               next_value=next_value_attr_1),
                  AttributeDef(name="attribute 2",
                               otf2_type=_otf2.TYPE_INT16,
                               next_value=next_value_attr_2)]


def add_attributes(attributes):
    if otf2_ATTRIBUTES:
        for a, attribute_def in enumerate(attribute_defs):
            _otf2.AttributeList_AddAttribute(attributes, a, attribute_def.type, attribute_def.next_value())


def record_for_all_locations(*_):
    return True


def record_for_location_0(rank, thread):
    return rank == 0 and thread == 0


# Main function
#
#  This tool generates a basic unified OTF2 archive to test the reader
#  components and tools.
#
#  @return                 Returns EXIT_SUCCESS if successful, EXIT_FAILURE
#                          if an error occures.
#
def main():
    archive_name = "TestTrace"

    # Array used to store metric values */
    metric_values = (_otf2.MetricValue * NUM_OF_METRICS)()

    # If you add new metrics please adapt also the array of metric types
    # and the routine to generate random metric values. */

    # Array of metric types */
    metric_types = [_otf2.TYPE_UINT64, _otf2.TYPE_DOUBLE, _otf2.TYPE_INT64,
                    _otf2.TYPE_DOUBLE, _otf2.TYPE_INT64]

    # Initialize accumulated metric */
    metric_values[0].unsigned_int = 0
    metric_values[1].floating_point = 12.5
    metric_values[2].signed_int = 1
    metric_values[3].floating_point = 10.0
    metric_values[4].unsigned_int = 1

    # Distribute metric members to its classes */
    metric_members_of_class_1 = [METRIC_A, METRIC_B]
    metric_members_of_class_2 = [METRIC_C]
    metric_members_of_class_3 = [METRIC_D]
    metric_members_of_class_4 = [METRIC_E]

    # Array of offsets needed to get correct type or values of a metric in corresponding arrays. */
    metric_classes_offsets = [0, 2, 3, 4, 4]

    # The collection of metric classes of instances for which we write metric events */
    class MetricWriteParameter:
        def __init__(self, id_, offset, count, record_for_location):
            self.id = id_
            self.metric_types = metric_types[offset:offset + count]
            self.metric_values = metric_values[offset:offset + count]
            self.record_for_location = record_for_location

    # FIXME missing class 4
    metric_write_parameters = [MetricWriteParameter(METRIC_CLASS_1, metric_classes_offsets[0],
                                                    len(metric_members_of_class_1),
                                                    record_for_all_locations),
                               MetricWriteParameter(METRIC_CLASS_2, metric_classes_offsets[1],
                                                    len(metric_members_of_class_2),
                                                    record_for_all_locations),
                               MetricWriteParameter(METRIC_CLASS_3, metric_classes_offsets[2],
                                                    len(metric_members_of_class_3),
                                                    record_for_all_locations),
                               MetricWriteParameter(METRIC_INSTANCE_1, metric_classes_offsets[3],
                                                    len(metric_members_of_class_4),
                                                    record_for_location_0)]

    def create_new_metric_values():
        metric_values[0].unsigned_int += 1
        metric_values[1].floating_point += 12.5
        metric_values[2].unsigned_int = -metric_values[2].unsigned_int
        metric_values[3].floating_point += 10.0
        metric_values[4].unsigned_int += 1

    def write_metrics(new, metric_time, metric_rank, metric_thread):
        if new:
            create_new_metric_values()

        for metric_write_parameter in metric_write_parameters:
            if metric_write_parameter.record_for_location(metric_rank, metric_thread):
                _otf2.EvtWriter_Metric(evt_writer, None, metric_time,
                                       metric_write_parameter.id,
                                       metric_write_parameter.metric_types,
                                       metric_write_parameter.metric_values)

    # Calculate number of MPI ranks and OpenMP threads. */
    num_ranks = 1
    num_threads = 1

    if otf2_MPI:
        num_ranks = otf2_LOCATIONS
        archive_path = "otf2_trace_gen_mpi_{}_{}{}".format(num_ranks, otf2_EVENTS,
                                                           "_attr" if otf2_ATTRIBUTES else "")
    elif otf2_OPENMP:
        num_threads = otf2_LOCATIONS
        archive_path = "otf2_trace_gen_omp_{}_{}{}".format(num_threads, otf2_EVENTS,
                                                           "_attr" if otf2_ATTRIBUTES else "")
    elif otf2_HYBRID:
        if (otf2_LOCATIONS % 4) == 0:
            num_threads = 4
        elif (otf2_LOCATIONS % 3) == 0:
            num_threads = 3
        elif (otf2_LOCATIONS % 2) == 0:
            num_threads = 2
        else:
            # FIXME shadow
            for num_threads in range(5, otf2_LOCATIONS):
                if (otf2_LOCATIONS % num_threads) == 0:
                    break

        num_ranks = otf2_LOCATIONS // num_threads
        archive_path = "otf2_trace_gen_mpi_omp_{}x{}_{}{}".format(num_ranks, num_threads, otf2_EVENTS,
                                                                  "_attr" if otf2_ATTRIBUTES else "")
    else:
        raise RuntimeError("No paradigm set")

    print("\nGenerating an OTF2 test archive")
    print("with {} rank(s) and {} thread(s) each ({} locations total)"
          .format(num_ranks, num_threads, otf2_LOCATIONS))
    print("to {}/{}.otf2\n".format(archive_path, archive_name))

    if os.path.isdir(archive_path):
        for i in itertools.count():
            archive_path_backup = "{}_bk_{}".format(archive_path, i)
            if not os.path.exists(archive_path_backup):
                os.rename(archive_path, archive_path_backup)
                break

    # Create new archive handle. */
    archive = _otf2.Archive_Open(archive_path, archive_name, _otf2.FILEMODE_WRITE,
                                 1024 * 1024, 4 * 1024 * 1024,
                                 _otf2.SUBSTRATE_POSIX, _otf2.COMPRESSION_NONE)

    flush_callbacks = _otf2.FlushCallbacks(pre_flush=pre_flush, post_flush=post_flush)
    _otf2.Archive_SetFlushCallbacks(archive, flush_callbacks, None)

    # Set description, and creator. */
    _otf2.Archive_SetSerialCollectiveCallbacks(archive)
    _otf2.Archive_SetDescription(archive, "Simple OTF2 trace archive to valid reader components")
    _otf2.Archive_SetCreator(archive, "otf2_trace_gen")

    # Generate location IDs. Just to have non-consecutive location IDs. */
    locations = []
    mpi_ranks = []
    initial_threads = []
    for rank in range(num_ranks):
        for thread in range(num_threads):
            locations.append((rank << 16) + thread)
        mpi_ranks.append(rank)
        initial_threads.append(rank << 16)

    _otf2.Archive_OpenEvtFiles(archive)

    # Create local event writer. */
    # Create an attribute list. */
    attributes = None
    if otf2_ATTRIBUTES:
        attributes = _otf2.AttributeList_New()

    t = get_time()
    for rank in range(num_ranks):
        evt_writer = _otf2.Archive_GetEvtWriter(archive, initial_threads[rank])

        # Enter main. */
        write_metrics(True, t, rank, 0)
        add_attributes(attributes)
        _otf2.EvtWriter_Enter(evt_writer, attributes, t, REGION_MAIN)

    for event in range(otf2_EVENTS):
        t = get_time()
        # FIXME - this makes no sense
        if not otf2_OPENMP:
            for rank in range(num_ranks):
                evt_writer = _otf2.Archive_GetEvtWriter(archive, initial_threads[rank])

                # Write enter MPI_Send. */
                write_metrics(True, t, rank, 0)
                add_attributes(attributes)
                _otf2.EvtWriter_Enter(evt_writer, attributes, t, REGION_MPI_SEND)
                _otf2.EvtWriter_MpiSend(evt_writer, None, t, (rank + 1) % num_ranks,
                                        rank, 0, 1024)

                t = get_time()

                # Write leave MPI_Send. */
                write_metrics(True, t, rank, 0)
                add_attributes(attributes)
                _otf2.EvtWriter_Leave(evt_writer, attributes, t, REGION_MPI_SEND)

                # Immediately enter the MPI_Recv */
                t = get_time()

                # Write enter MPI_Recv. */
                write_metrics(True, t, rank, 0)
                add_attributes(attributes)
                _otf2.EvtWriter_Enter(evt_writer, attributes, t, REGION_MPI_RECV)

        if not otf2_OPENMP:
            for rank in range(num_ranks):
                evt_writer = _otf2.Archive_GetEvtWriter(archive, initial_threads[rank])

                _otf2.EvtWriter_MpiRecv(evt_writer, None, t,
                                        (rank + num_ranks - 1) % num_ranks,
                                        rank, 0, 1024)

                # Write enter MPI_Recv. */
                write_metrics(True, t, rank, 0)
                add_attributes(attributes)
                _otf2.EvtWriter_Leave(evt_writer, attributes, t, REGION_MPI_RECV)

                # MPI Barrier. */
                t = get_time()

                write_metrics(True, t, rank, 0)
                add_attributes(attributes)
                _otf2.EvtWriter_Enter(evt_writer, attributes, t, REGION_MPI_BARRIER)

                _otf2.EvtWriter_MpiCollectiveBegin(evt_writer, None, t)

        # End all started MPI_Barrier */
        t = get_time()
        if not otf2_OPENMP:
            for rank in range(num_ranks):
                evt_writer = _otf2.Archive_GetEvtWriter(archive, initial_threads[rank])
                _otf2.EvtWriter_MpiCollectiveEnd(evt_writer, None, t,
                                                 _otf2.COLLECTIVE_OP_BARRIER, rank,
                                                 _otf2.COLLECTIVE_ROOT_NONE, 0, 0)

                write_metrics(True, t, rank, 0)
                add_attributes(attributes)
                _otf2.EvtWriter_Leave(evt_writer, attributes, t, REGION_MPI_BARRIER)

        # If in OpenMP or Hybrid mode. */
        if not otf2_MPI:
            t_enter = get_time()
            t_leave = get_time()

            for rank in range(num_ranks):
                evt_writer = _otf2.Archive_GetEvtWriter(archive, initial_threads[rank])
                # At this point the initial thread creates a team of several parallel threads */
                _otf2.EvtWriter_OmpFork(evt_writer, attributes, t_enter, num_threads)

            for rank in range(num_ranks):
                for thread in range(num_threads):
                    evt_writer = _otf2.Archive_GetEvtWriter(archive,
                                                            locations[num_threads * rank + thread])

                    # Enter OpenMP parallel  loop (outer region). */
                    write_metrics(True, t_enter, rank, thread)
                    add_attributes(attributes)
                    _otf2.EvtWriter_Enter(evt_writer, attributes, t_enter, REGION_OPENMP_PARALLEL)

                    # Enter OpenMP parallel loop  (inner region). */
                    write_metrics(False, t_enter, rank, thread)
                    add_attributes(attributes)
                    _otf2.EvtWriter_Enter(evt_writer, attributes, t_enter, REGION_OPENMP_LOOP)

                    # Enter OpenMP implicit barrier. */
                    write_metrics(False, t_enter, rank, thread)
                    add_attributes(attributes)
                    _otf2.EvtWriter_Enter(evt_writer, attributes, t_enter, REGION_OPENMP_IMPLICIT_BARRIER)

            for rank in range(num_ranks):
                for thread in range(num_threads):
                    evt_writer = _otf2.Archive_GetEvtWriter(archive,
                                                            locations[num_threads * rank + thread])

                    # Leave OpenMP implicit barrier. */
                    write_metrics(True, t_leave, rank, thread)
                    add_attributes(attributes)
                    _otf2.EvtWriter_Leave(evt_writer, attributes, t_leave, REGION_OPENMP_IMPLICIT_BARRIER)

                    # Leave OpenMP parallel loop (inner region). */
                    write_metrics(False, t_leave, rank, thread)
                    add_attributes(attributes)
                    _otf2.EvtWriter_Leave(evt_writer, attributes, t_leave, REGION_OPENMP_LOOP)

                    # Leave OpenMP parallel loop (outer region). */
                    write_metrics(False, t_leave, rank, thread)
                    add_attributes(attributes)
                    _otf2.EvtWriter_Leave(evt_writer, attributes, t_leave, REGION_OPENMP_PARALLEL)

            for rank in range(num_ranks):
                evt_writer = _otf2.Archive_GetEvtWriter(archive, initial_threads[rank])
                # All threads have finished execution of statements in parallel region.
                # Now they synchronize and terminates. Only the initial thread continues execution. */
                t_leave = get_time()
                _otf2.EvtWriter_OmpJoin(evt_writer, attributes, t_leave)

        if not otf2_OPENMP:
            for rank in range(num_ranks):
                evt_writer = _otf2.Archive_GetEvtWriter(archive, initial_threads[rank])

                t = get_time()

                write_metrics(True, t, rank, 0)
                add_attributes(attributes)
                _otf2.EvtWriter_Enter(evt_writer, attributes, t, REGION_MPI_BARRIER)

                _otf2.EvtWriter_MpiCollectiveBegin(evt_writer, None, t)

        # End all started MPI_Barrier */
        t = get_time()
        if not otf2_OPENMP:
            for rank in range(num_ranks):
                evt_writer = _otf2.Archive_GetEvtWriter(archive, initial_threads[rank])

                _otf2.EvtWriter_MpiCollectiveEnd(evt_writer, None, t,
                                                 _otf2.COLLECTIVE_OP_BARRIER,
                                                 rank, _otf2.COLLECTIVE_ROOT_NONE,
                                                 0, 0)

                write_metrics(True, t, rank, 0)
                add_attributes(attributes)
                _otf2.EvtWriter_Leave(evt_writer, attributes, t, REGION_MPI_BARRIER)

    t = get_time()
    for rank in range(num_ranks):
        evt_writer = _otf2.Archive_GetEvtWriter(archive, initial_threads[rank])

        # Leave main. */
        write_metrics(True, t, rank, 0)
        add_attributes(attributes)
        _otf2.EvtWriter_Leave(evt_writer, attributes, t, REGION_MAIN)

    # last timestamp */
    t = get_time()
    number_of_events_per_location = []
    for rank in range(num_ranks):
        for thread in range(num_threads):
            evt_writer = _otf2.Archive_GetEvtWriter(archive,
                                                    locations[num_threads * rank + thread])

            n = _otf2.EvtWriter_GetNumberOfEvents(evt_writer)
            number_of_events_per_location.append(n)

            _otf2.Archive_CloseEvtWriter(archive, evt_writer)

    _otf2.Archive_CloseEvtFiles(archive)

    # Write the global and local definitions. */
    glob_def_writer = _otf2.Archive_GetGlobalDefWriter(archive)

    # Write the timerange of the trace */
    _otf2.GlobalDefWriter_WriteClockProperties(glob_def_writer, 1, 0, t, _otf2.UNDEFINED_TIMESTAMP)

    _otf2.GlobalDefWriter_WriteString(glob_def_writer, STRING_EMPTY, "")

    string = STRING_EMPTY + 1

    # Write paradigm definitions */
    if otf2_MPI or otf2_HYBRID:
        _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "MPI")

        _otf2.GlobalDefWriter_WriteParadigm(glob_def_writer,
                                            _otf2.PARADIGM_MPI,
                                            string,
                                            _otf2.PARADIGM_CLASS_PROCESS)

        string += 1

        _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "MPI_COMM_${id}")

        value = _otf2.AttributeValue()
        value.stringRef = string
        string += 1
        _otf2.GlobalDefWriter_WriteParadigmProperty(glob_def_writer,
                                                    _otf2.PARADIGM_MPI,
                                                    _otf2.PARADIGM_PROPERTY_COMM_NAME_TEMPLATE,
                                                    _otf2.TYPE_STRING,
                                                    value)

        _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "MPI_WIN_${id}")

        value.stringRef = string
        string += 1
        _otf2.GlobalDefWriter_WriteParadigmProperty(glob_def_writer,
                                                    _otf2.PARADIGM_MPI,
                                                    _otf2.PARADIGM_PROPERTY_RMA_WIN_NAME_TEMPLATE,
                                                    _otf2.TYPE_STRING,
                                                    value)

        (type_, value) = _otf2.AttributeValue_SetBoolean(_otf2.FALSE)
        _otf2.GlobalDefWriter_WriteParadigmProperty(glob_def_writer,
                                                    _otf2.PARADIGM_OPENMP,
                                                    _otf2.PARADIGM_PROPERTY_RMA_ONLY,
                                                    type_,
                                                    value)

    if otf2_OPENMP or otf2_HYBRID:
        _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "OpenMP")

        _otf2.GlobalDefWriter_WriteParadigm(glob_def_writer,
                                            _otf2.PARADIGM_OPENMP,
                                            string,
                                            _otf2.PARADIGM_CLASS_THREAD_FORK_JOIN)
        string += 1

        _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "OMP Thread Team ${id}")

        value = _otf2.AttributeValue()
        value.stringRef = string
        string += 1
        # FIXME seems like a bad idea to write a hardcoded number 4 here
        _otf2.GlobalDefWriter_WriteParadigmProperty(glob_def_writer,
                                                    _otf2.PARADIGM_OPENMP,
                                                    4, _otf2.TYPE_STRING,
                                                    value)

    # Write system tree definitions. */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "Rack 0")

    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string + 1, "Rack")

    _otf2.GlobalDefWriter_WriteSystemTreeNode(glob_def_writer, 0,
                                              string, string + 1,
                                              _otf2.UNDEFINED_SYSTEM_TREE_NODE)

    string += 2

    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "Blade")

    string_blade = string
    string += 1

    for rank in range(num_ranks // 2):
        name_buffer = "Blade #{}".format(rank)

        _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, name_buffer)

        _otf2.GlobalDefWriter_WriteSystemTreeNode(glob_def_writer, rank + 1,
                                                  string, string_blade, 0)
        string += 1

    # Write location group and location definitions. */
    for rank in range(num_ranks):
        name_buffer = "Process {}".format(rank)
        _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, name_buffer)

        _otf2.GlobalDefWriter_WriteLocationGroup(glob_def_writer, rank,
                                                 string,
                                                 _otf2.LOCATION_GROUP_TYPE_PROCESS,
                                                 (rank // 2) + 1, _otf2.UNDEFINED_LOCATION_GROUP)
        string += 1

        for thread in range(num_threads):
            name_buffer = "Thread {}.{}".format(rank, thread)
            _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, name_buffer)

            _otf2.GlobalDefWriter_WriteLocation(glob_def_writer,
                                                locations[rank * num_threads + thread],
                                                string,
                                                _otf2.LOCATION_TYPE_CPU_THREAD,
                                                number_of_events_per_location[rank * num_threads + thread],
                                                rank)
            string += 1

    # Write region definition for "main". */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string,
                                      "main (mangled name)")
    string += 1
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string,
                                      "main (demangled name)")

    _otf2.GlobalDefWriter_WriteRegion(glob_def_writer, REGION_MAIN,
                                      string, string - 1,
                                      STRING_EMPTY,
                                      _otf2.REGION_ROLE_UNKNOWN,
                                      _otf2.PARADIGM_UNKNOWN,
                                      _otf2.REGION_FLAG_NONE,
                                      STRING_EMPTY, 0, 0)
    string += 1

    # Write region definition for an all-threads function. */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string,
                                      "!$omp parallelfor")
    _otf2.GlobalDefWriter_WriteRegion(glob_def_writer,
                                      REGION_OPENMP_PARALLEL,
                                      string, string, STRING_EMPTY,
                                      _otf2.REGION_ROLE_UNKNOWN,
                                      _otf2.PARADIGM_UNKNOWN,
                                      _otf2.REGION_FLAG_NONE,
                                      STRING_EMPTY, 0, 0)

    _otf2.GlobalDefWriter_WriteRegion(glob_def_writer, REGION_OPENMP_LOOP,
                                      string, string, STRING_EMPTY,
                                      _otf2.REGION_ROLE_UNKNOWN,
                                      _otf2.PARADIGM_UNKNOWN,
                                      _otf2.REGION_FLAG_NONE,
                                      STRING_EMPTY, 0, 0)
    string += 1

    # Write region definition for OpenMP implicit barrier. */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string,
                                      "!$omp implicit barrier")

    _otf2.GlobalDefWriter_WriteRegion(glob_def_writer,
                                      REGION_OPENMP_IMPLICIT_BARRIER,
                                      string, string, STRING_EMPTY,
                                      _otf2.REGION_ROLE_UNKNOWN,
                                      _otf2.PARADIGM_UNKNOWN,
                                      _otf2.REGION_FLAG_NONE,
                                      STRING_EMPTY, 0, 0)
    string += 1

    # Write region definition for "MPI_Send". */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "MPI_Send")

    _otf2.GlobalDefWriter_WriteRegion(glob_def_writer, REGION_MPI_SEND,
                                      string, string, STRING_EMPTY,
                                      _otf2.REGION_ROLE_UNKNOWN,
                                      _otf2.PARADIGM_UNKNOWN,
                                      _otf2.REGION_FLAG_NONE,
                                      STRING_EMPTY, 0, 0)
    string += 1

    # Write region definition for "MPI_Recv". */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "MPI_Recv")

    _otf2.GlobalDefWriter_WriteRegion(glob_def_writer, REGION_MPI_RECV,
                                      string, string, STRING_EMPTY,
                                      _otf2.REGION_ROLE_UNKNOWN,
                                      _otf2.PARADIGM_UNKNOWN,
                                      _otf2.REGION_FLAG_NONE,
                                      STRING_EMPTY, 0, 0)
    string += 1

    # Write region definition for "MPI_Barrier". */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "MPI_Barrier")

    _otf2.GlobalDefWriter_WriteRegion(glob_def_writer, REGION_MPI_BARRIER,
                                      string, string, STRING_EMPTY,
                                      _otf2.REGION_ROLE_UNKNOWN,
                                      _otf2.PARADIGM_UNKNOWN,
                                      _otf2.REGION_FLAG_NONE,
                                      STRING_EMPTY, 0, 0)
    string += 1

    # Define the list of locations which are mpi ranks */
    _otf2.GlobalDefWriter_WriteGroup(glob_def_writer, GROUP_MPI_LOCATIONS, STRING_EMPTY,
                                     _otf2.GROUP_TYPE_COMM_LOCATIONS, _otf2.PARADIGM_MPI,
                                     _otf2.GROUP_FLAG_NONE, initial_threads)

    # Define the MPI group for MPI_COMM_WORLD */
    _otf2.GlobalDefWriter_WriteGroup(glob_def_writer, GROUP_MPI_COMM_WORLD, STRING_EMPTY,
                                     _otf2.GROUP_TYPE_COMM_GROUP, _otf2.PARADIGM_MPI,
                                     _otf2.GROUP_FLAG_NONE, mpi_ranks)

    # Define the MPI group for COMM_SELF like communicators */
    _otf2.GlobalDefWriter_WriteGroup(glob_def_writer, GROUP_MPI_COMM_SELF, STRING_EMPTY, _otf2.GROUP_TYPE_COMM_SELF,
                                     _otf2.PARADIGM_MPI, _otf2.GROUP_FLAG_NONE, [])

    # Define a group that includes all locations. It will be used by metric instance. */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "All locations")

    _otf2.GlobalDefWriter_WriteGroup(glob_def_writer, GROUP_ALL_LOCATIONS, string, _otf2.GROUP_TYPE_LOCATIONS,
                                     _otf2.PARADIGM_UNKNOWN, _otf2.GROUP_FLAG_NONE, locations)
    string += 1

    # Define MPI_COMM_WORD */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "My MPI_COMM_WORLD communicator")
    _otf2.GlobalDefWriter_WriteComm(glob_def_writer, MPI_COMM_MPI_COMM_WORLD,
                                    string, GROUP_MPI_COMM_WORLD, _otf2.UNDEFINED_COMM,
                                    _otf2.COMM_FLAG_NONE)
    string += 1

    # Define MPI_COMM_SELF */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "My MPI_COMM_SELF communicator")
    _otf2.GlobalDefWriter_WriteComm(glob_def_writer, MPI_COMM_MPI_COMM_SELF,
                                    string, GROUP_MPI_COMM_SELF, _otf2.UNDEFINED_COMM,
                                    _otf2.COMM_FLAG_NONE)
    string += 1

    # Write member definition of first metric */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "Metric_A")
    string += 1

    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "FLOPS")
    _otf2.GlobalDefWriter_WriteMetricMember(glob_def_writer, METRIC_A, string - 1, STRING_EMPTY,
                                            _otf2.METRIC_TYPE_PAPI, _otf2.METRIC_ACCUMULATED_START,
                                            metric_types[0], _otf2.BASE_DECIMAL, 0, string)
    string += 1

    # Write member definition of second metric */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "Metric_B")
    string += 1
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "s")

    _otf2.GlobalDefWriter_WriteMetricMember(glob_def_writer, METRIC_B, string - 1,
                                            STRING_EMPTY, _otf2.METRIC_TYPE_OTHER,
                                            _otf2.METRIC_ABSOLUTE_POINT, metric_types[1],
                                            _otf2.BASE_DECIMAL, -6, string)
    string += 1

    # Write member definition of third metric */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "Metric_C")
    string += 1
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "apples")

    _otf2.GlobalDefWriter_WriteMetricMember(glob_def_writer, METRIC_C, string - 1, STRING_EMPTY,
                                            _otf2.METRIC_TYPE_OTHER, _otf2.METRIC_RELATIVE_LAST,
                                            metric_types[2], _otf2.BASE_DECIMAL, 0, string)
    string += 1

    # Write member definition of fourth metric */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "Metric_D")
    string += 1
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "pears")

    _otf2.GlobalDefWriter_WriteMetricMember(glob_def_writer, METRIC_D, string - 1, STRING_EMPTY,
                                            _otf2.METRIC_TYPE_OTHER, _otf2.METRIC_ABSOLUTE_POINT,
                                            metric_types[3], _otf2.BASE_DECIMAL, 0, string)
    string += 1

    # Write member definition of fifth metric */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "Metric_E")
    string += 1
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "cherries")

    _otf2.GlobalDefWriter_WriteMetricMember(glob_def_writer, METRIC_E, string - 1, STRING_EMPTY,
                                            _otf2.METRIC_TYPE_OTHER, _otf2.METRIC_ABSOLUTE_POINT,
                                            metric_types[4], _otf2.BASE_DECIMAL, 0, string)
    string += 1

    # Define metric classes and instances */
    _otf2.GlobalDefWriter_WriteMetricClass(glob_def_writer, METRIC_CLASS_1, metric_members_of_class_1,
                                           _otf2.METRIC_SYNCHRONOUS_STRICT, _otf2.RECORDER_KIND_CPU)
    _otf2.GlobalDefWriter_WriteMetricClass(glob_def_writer, METRIC_CLASS_2, metric_members_of_class_2,
                                           _otf2.METRIC_SYNCHRONOUS_STRICT, _otf2.RECORDER_KIND_CPU)

    _otf2.GlobalDefWriter_WriteMetricClass(glob_def_writer, METRIC_CLASS_3, metric_members_of_class_3,
                                           _otf2.METRIC_SYNCHRONOUS_STRICT, _otf2.RECORDER_KIND_CPU)

    _otf2.GlobalDefWriter_WriteMetricClass(glob_def_writer, METRIC_CLASS_4, metric_members_of_class_4,
                                           _otf2.METRIC_ASYNCHRONOUS, _otf2.RECORDER_KIND_ABSTRACT)

    # We can define an instance with differing recorder and scope.
    # In this example metric values are valid for all locationa
    # but are recorded on first location. */
    _otf2.GlobalDefWriter_WriteMetricInstance(glob_def_writer, METRIC_INSTANCE_1, METRIC_CLASS_4, locations[0],
                                              _otf2.SCOPE_GROUP, GROUP_ALL_LOCATIONS)

    # write attribute defs */
    if otf2_ATTRIBUTES:
        for a, attribute_def in enumerate(attribute_defs):
            _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, attribute_def.name)
            _otf2.GlobalDefWriter_WriteAttribute(glob_def_writer, a, string, STRING_EMPTY, attribute_def.type)
            string += 1

    # create a sub communicator of MPI_COMM_WORLD and embed an hyper cube. */
    num_cube_ranks = gpot(num_ranks)
    cube_ranks = list(range(num_cube_ranks))

    # Define the MPI group for the hyper cube */
    _otf2.GlobalDefWriter_WriteGroup(glob_def_writer, GROUP_MPI_COMM_CUBE, STRING_EMPTY,
                                     _otf2.GROUP_TYPE_COMM_GROUP, _otf2.PARADIGM_MPI,
                                     _otf2.GROUP_FLAG_NONE, cube_ranks)

    # Define MPI_COMM_MPI_COMM_CUBE */
    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "MPI_COMM_CUBE")
    _otf2.GlobalDefWriter_WriteComm(glob_def_writer, MPI_COMM_MPI_COMM_CUBE, string,
                                    GROUP_MPI_COMM_WORLD, MPI_COMM_MPI_COMM_WORLD,
                                    _otf2.COMM_FLAG_NONE)

    string += 1

    num_cube_dims = 0
    cube_dim = 1
    while cube_dim < num_cube_ranks:
        name = "Cube {}".format(cube_dim)
        _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, name)
        _otf2.GlobalDefWriter_WriteCartDimension(glob_def_writer, num_cube_dims, string, 2, _otf2.CART_PERIODIC_FALSE)
        string += 1
        num_cube_dims += 1
        cube_dim <<= 1

    cube_dimensions = [_otf2.CartDimensionRef(cube_dim) for cube_dim in range(num_cube_dims)]

    _otf2.GlobalDefWriter_WriteString(glob_def_writer, string, "Hyper Cube")
    _otf2.GlobalDefWriter_WriteCartTopology(glob_def_writer, 0, string, MPI_COMM_MPI_COMM_CUBE, cube_dimensions)
    string += 1

    for rank in range(num_cube_ranks):
        cube_coordinates = [1 if rank & (2 ** cube_dim) else 0 for cube_dim in range(num_cube_dims)]
        _otf2.GlobalDefWriter_WriteCartCoordinate(glob_def_writer, 0, rank, cube_coordinates)

    _otf2.Archive_CloseGlobalDefWriter(archive, glob_def_writer)

    # write local mappings for MPI communicators and metrics */
    # write local mappings for metrics, this is an identity map, just to write
    # something out

    _otf2.Archive_OpenDefFiles(archive)

    metric_map = _otf2.IdMap_Create(_otf2.ID_MAP_DENSE, NUM_OF_CLASSES)
    for c in range(NUM_OF_CLASSES):
        _otf2.IdMap_AddIdPair(metric_map, c, c)

    for rank in range(num_ranks):
        mpi_comm_map = _otf2.IdMap_Create(_otf2.ID_MAP_SPARSE, 2)

        # Each location uses its rank as the communicator id which maps to the global 0 */
        _otf2.IdMap_AddIdPair(mpi_comm_map, rank, MPI_COMM_MPI_COMM_WORLD)
        _otf2.IdMap_AddIdPair(mpi_comm_map, rank + num_ranks, MPI_COMM_MPI_COMM_SELF)

        for thread in range(num_threads):
            # Open a definition writer, so the file is created. */
            def_writer = _otf2.Archive_GetDefWriter(archive, locations[num_threads * rank + thread])

            # Write metric mappings to local definitions. */
            _otf2.DefWriter_WriteMappingTable(def_writer, _otf2.MAPPING_METRIC, metric_map)

            if otf2_MPI or otf2_HYBRID:
                # Write MPI Comm mappings to local definitions. */
                _otf2.DefWriter_WriteMappingTable(def_writer, _otf2.MAPPING_COMM, mpi_comm_map)

            # Write clock offsets to local definitions. */
            _otf2.DefWriter_WriteClockOffset(def_writer, 0, 0, 0.0)
            _otf2.DefWriter_WriteClockOffset(def_writer, t, 0, 0.0)

            _otf2.Archive_CloseDefWriter(archive, def_writer)

        _otf2.IdMap_Free(mpi_comm_map)
    _otf2.IdMap_Free(metric_map)

    _otf2.Archive_CloseDefFiles(archive)

    # Finalize. */
    _otf2.Archive_Close(archive)

    _otf2.AttributeList_Delete(attributes)

    print("Done\n")


# ___ Implementation of static functions ___________________________________ */

# @internal
#  @brief Get command line parameters.
#
#  Parses command line parameters and checks for their existence.
#  Prints help for parameters '-h' or '--help'.
#
#  @param argc             Programs argument counter.
#  @param argv             Programs argument values.
def otf2_get_parameters():
    parser = argparse.ArgumentParser(description="Generate a basic unified OTF2 archive")
    parser.add_argument("-e", "--events", type=int, default=1, help="Approx. number of events to generate per location")
    parser.add_argument("-l", "--locations", type=int, default=4,
                        help="Number of locations to generate (Should be small)")
    parser.add_argument("-V", "--version", action="store_true")
    parser.add_argument("--mpi", action="store_true", help="MPI-only mode (default)")
    parser.add_argument("--openmp", action="store_true", help="OpenMP-only mode")
    parser.add_argument("--hybrid", action="store_true", help="Hybrid mode")
    parser.add_argument("--attributes", action="store_true", help="Add attributes to enter/leave events")
    args = parser.parse_args()

    if args.version:
        print("otf2-trace-gen: version {}".format(_otf2.VERSION))
        exit(0)

    global otf2_MPI, otf2_OPENMP, otf2_HYBRID, otf2_ATTRIBUTES, otf2_LOCATIONS, otf2_EVENTS
    otf2_MPI = False
    otf2_OPENMP = False
    otf2_HYBRID = False
    if args.mpi:
        otf2_MPI = True
    elif args.openmp:
        otf2_OPENMP = True
    elif args.hybrid:
        otf2_HYBRID = True
    else:
        otf2_MPI = True

    otf2_ATTRIBUTES = args.attributes
    otf2_LOCATIONS = args.locations
    otf2_EVENTS = args.events


def otf2_set_parameters(paradigm="mpi", attributes=False, locations=4, events=1):
    global otf2_MPI, otf2_OPENMP, otf2_HYBRID, otf2_ATTRIBUTES, otf2_LOCATIONS, otf2_EVENTS
    otf2_MPI = False
    otf2_OPENMP = False
    otf2_HYBRID = False
    if paradigm == 'mpi':
        otf2_MPI = True
    elif paradigm == 'openmp':
        otf2_OPENMP = True
    elif paradigm == 'hybrid':
        otf2_HYBRID = True
    otf2_ATTRIBUTES = attributes
    otf2_LOCATIONS = locations
    otf2_EVENTS = events


if __name__ == '__main__':
    otf2_get_parameters()
    main()



import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import LocationRef, TimeStamp
from .EvtReaderCallbacks import EvtReaderCallbacks


# forward declaration
class EvtReader(ctypes.Structure):
    pass

def EvtReader_GetLocationID(reader):
    c_GetLocationID = conf.lib.OTF2_EvtReader_GetLocationID
    c_GetLocationID.argtypes = [ ctypes.POINTER(EvtReader), ctypes.POINTER(LocationRef) ]
    c_GetLocationID.restype = ErrorCode
    c_GetLocationID.errcheck = HandleErrorCode
    location = LocationRef()
    c_GetLocationID(reader, ctypes.byref(location))
    return location.value

def EvtReader_GetPos(reader):
    c_GetPos = conf.lib.OTF2_EvtReader_GetPos
    c_GetPos.argtypes = [ ctypes.POINTER(EvtReader), ctypes.POINTER(ctypes.c_uint64) ]
    c_GetPos.restype = ErrorCode
    c_GetPos.errcheck = HandleErrorCode
    position = ctypes.c_uint64()
    c_GetPos(reader, ctypes.byref(position))
    return position.value

def EvtReader_SetCallbacks(reader, callbacks, userData):
    c_SetCallbacks = conf.lib.OTF2_EvtReader_SetCallbacks
    c_SetCallbacks.argtypes = [ ctypes.POINTER(EvtReader), ctypes.POINTER(EvtReaderCallbacks), ctypes.py_object ]
    c_SetCallbacks.restype = ErrorCode
    c_SetCallbacks.errcheck = HandleErrorCode
    c_SetCallbacks(reader, callbacks, ctypes.py_object(userData))

def EvtReader_Seek(reader, position):
    c_Seek = conf.lib.OTF2_EvtReader_Seek
    c_Seek.argtypes = [ ctypes.POINTER(EvtReader), ctypes.c_uint64 ]
    c_Seek.restype = ErrorCode
    c_Seek.errcheck = HandleErrorCode
    c_Seek(reader, position)

def EvtReader_ReadEvents(reader, recordsToRead):
    c_ReadEvents = conf.lib.OTF2_EvtReader_ReadEvents
    c_ReadEvents.argtypes = [ ctypes.POINTER(EvtReader), ctypes.c_uint64, ctypes.POINTER(ctypes.c_uint64) ]
    c_ReadEvents.restype = ErrorCode
    c_ReadEvents.errcheck = HandleErrorCode
    recordsRead = ctypes.c_uint64()
    c_ReadEvents(reader, recordsToRead, ctypes.byref(recordsRead))
    return recordsRead.value

def EvtReader_TimeStampRewrite(reader, time):
    c_TimeStampRewrite = conf.lib.OTF2_EvtReader_TimeStampRewrite
    c_TimeStampRewrite.argtypes = [ ctypes.POINTER(EvtReader), TimeStamp ]
    c_TimeStampRewrite.restype = ErrorCode
    c_TimeStampRewrite.errcheck = HandleErrorCode
    c_TimeStampRewrite(reader, time)

def EvtReader_ReadEventsBackward(reader, recordsToRead):
    c_ReadEventsBackward = conf.lib.OTF2_EvtReader_ReadEventsBackward
    c_ReadEventsBackward.argtypes = [ ctypes.POINTER(EvtReader), ctypes.c_uint64, ctypes.POINTER(ctypes.c_uint64) ]
    c_ReadEventsBackward.restype = ErrorCode
    c_ReadEventsBackward.errcheck = HandleErrorCode
    recordsRead = ctypes.c_uint64()
    c_ReadEventsBackward(reader, recordsToRead, ctypes.byref(recordsRead))
    return recordsRead.value

def EvtReader_ApplyMappingTables(reader, action):
    c_ApplyMappingTables = conf.lib.OTF2_EvtReader_ApplyMappingTables
    c_ApplyMappingTables.argtypes = [ ctypes.POINTER(EvtReader), ctypes.c_int ]
    c_ApplyMappingTables.restype = ErrorCode
    c_ApplyMappingTables.errcheck = HandleErrorCode
    c_ApplyMappingTables(reader, action)

def EvtReader_ApplyClockOffsets(reader, action):
    c_ApplyClockOffsets = conf.lib.OTF2_EvtReader_ApplyClockOffsets
    c_ApplyClockOffsets.argtypes = [ ctypes.POINTER(EvtReader), ctypes.c_int ]
    c_ApplyClockOffsets.restype = ErrorCode
    c_ApplyClockOffsets.errcheck = HandleErrorCode
    c_ApplyClockOffsets(reader, action)

__all__ = [
    'EvtReader',
    'EvtReader_ApplyClockOffsets',
    'EvtReader_ApplyMappingTables',
    'EvtReader_GetLocationID',
    'EvtReader_GetPos',
    'EvtReader_ReadEvents',
    'EvtReader_ReadEventsBackward',
    'EvtReader_Seek',
    'EvtReader_SetCallbacks',
    'EvtReader_TimeStampRewrite'
]



#!/usr/bin/env python

import os
import os.path
import unittest
from shutil import rmtree
from tempfile import mkdtemp

from _otf2_trace_gen import main, otf2_set_parameters


class TestOTF2TraceGen(unittest.TestCase):
    def setUp(self):
        self.old_cwd = os.getcwd()
        self.tmp_dirname = mkdtemp(prefix=os.path.basename(os.path.abspath(__file__))[:-3] + '_tmp', dir=self.old_cwd)
        os.chdir(self.tmp_dirname)

    def tearDown(self):
        os.chdir(self.old_cwd)
        if os.getenv('KEEP_TEST_OUTPUT', '') != '':
            print(self.tmp_dirname)
        else:
            rmtree(self.tmp_dirname)

    def test_otf2_trace_gen_mpi(self):
        otf2_set_parameters(paradigm='mpi', locations=4, attributes=True, events=10)
        main()

    def test_otf2_trace_gen_openmp(self):
        otf2_set_parameters(paradigm='openmp', locations=6, attributes=False, events=20)
        main()

    def test_otf2_trace_gen_hybrid(self):
        otf2_set_parameters(paradigm='hybrid', locations=12, attributes=True, events=7)
        main()


if __name__ == '__main__':
    unittest.main()




''' Generated by OTF2 Template Engine '''

import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import *
from .AttributeValue import AttributeValue


class AttributeList(ctypes.Structure):
    pass

def AttributeList_New():
    c_New = conf.lib.OTF2_AttributeList_New
    c_New.argtypes = []
    c_New.restype = ctypes.POINTER(AttributeList)
    # NOTE: Do not errcheck here. This function returns a value, not an error code
    return c_New()

def AttributeList_Delete(attributeList):
    c_Delete = conf.lib.OTF2_AttributeList_Delete
    c_Delete.argtypes = [ ctypes.POINTER(AttributeList) ]
    c_Delete.restype = ErrorCode
    c_Delete.errcheck = HandleErrorCode
    c_Delete(attributeList)

def AttributeList_AddAttribute(attributeList, attribute, type, attributeValue):
    c_AddAttribute = conf.lib.OTF2_AttributeList_AddAttribute
    c_AddAttribute.argtypes = [ ctypes.POINTER(AttributeList), AttributeRef, Type, AttributeValue._arg_type ]
    c_AddAttribute.restype = ErrorCode
    c_AddAttribute.errcheck = HandleErrorCode
    c_AddAttribute(attributeList, attribute, type, attributeValue._arg_value)

@otf2 for entry in Type.entries if entry.type:
def AttributeList_Add@@entry.type.otf2@@(attributeList, attribute, @@entry.type.lower@@Value):
    c_Add@@entry.type.otf2@@ = conf.lib.OTF2_AttributeList_Add@@entry.type.otf2@@
    c_Add@@entry.type.otf2@@.argtypes = [ ctypes.POINTER(AttributeList), AttributeRef, @@entry.type.py_argtype@@ ]
    c_Add@@entry.type.otf2@@.restype = ErrorCode
    c_Add@@entry.type.otf2@@.errcheck = HandleErrorCode
    c_Add@@entry.type.otf2@@(attributeList, attribute, @@entry.type.lower@@Value)

@otf2 endfor
def AttributeList_AddString(attributeList, attribute, stringRef):
    c_AddString = conf.lib.OTF2_AttributeList_AddString
    c_AddString.argtypes = [ctypes.POINTER(AttributeList), AttributeRef, StringRef]
    c_AddString.restype = ErrorCode
    c_AddString.errcheck = HandleErrorCode
    c_AddString(attributeList, attribute, stringRef)

@otf2 for entry in MappingType.entries:
@otf2  with def = entry.def
def AttributeList_Add@@entry.def.ref_name@@Ref(attributeList, attribute, @@def.ref_lname@@Ref):
    c_Add@@entry.def.ref_name@@Ref = conf.lib.OTF2_AttributeList_Add@@entry.def.ref_name@@Ref
    c_Add@@entry.def.ref_name@@Ref.argtypes = [ ctypes.POINTER(AttributeList), AttributeRef, @@def.self_attr.type.py_argtype@@ ]
    c_Add@@entry.def.ref_name@@Ref.restype = ErrorCode
    c_Add@@entry.def.ref_name@@Ref.errcheck = HandleErrorCode
    c_Add@@entry.def.ref_name@@Ref(attributeList, attribute, @@def.ref_lname@@Ref)

@otf2 endwith
@otf2 endfor
def AttributeList_RemoveAttribute(attributeList, attribute):
    c_RemoveAttribute = conf.lib.OTF2_AttributeList_RemoveAttribute
    c_RemoveAttribute.argtypes = [ ctypes.POINTER(AttributeList), AttributeRef ]
    c_RemoveAttribute.restype = ErrorCode
    c_RemoveAttribute.errcheck = HandleErrorCode
    c_RemoveAttribute(attributeList, attribute)

def AttributeList_RemoveAllAttributes(attributeList):
    c_RemoveAllAttributes = conf.lib.OTF2_AttributeList_RemoveAllAttributes
    c_RemoveAllAttributes.argtypes = [ ctypes.POINTER(AttributeList) ]
    c_RemoveAllAttributes.restype = ErrorCode
    c_RemoveAllAttributes.errcheck = HandleErrorCode
    c_RemoveAllAttributes(attributeList)

def AttributeList_TestAttributeByID(attributeList, attribute):
    c_TestAttributeByID = conf.lib.OTF2_AttributeList_TestAttributeByID
    c_TestAttributeByID.argtypes = [ ctypes.POINTER(AttributeList), AttributeRef ]
    c_TestAttributeByID.restype = ctypes.c_bool
    return c_TestAttributeByID(attributeList, attribute)

def AttributeList_GetAttributeByID(attributeList, attribute, attributeType = Type(), attributeValue = AttributeValue()):
    c_GetAttributeByID = conf.lib.OTF2_AttributeList_GetAttributeByID
    c_GetAttributeByID.argtypes = [ ctypes.POINTER(AttributeList), AttributeRef, ctypes.POINTER(Type), ctypes.POINTER(AttributeValue) ]
    c_GetAttributeByID.restype = ErrorCode
    c_GetAttributeByID.errcheck = HandleErrorCode
    c_GetAttributeByID(attributeList, attribute, ctypes.byref(attributeType), ctypes.byref(attributeValue))
    return (attributeType, attributeValue)

@otf2 for entry in Type.entries if entry.type:
def AttributeList_Get@@entry.type.otf2@@(attributeList, attribute):
    c_Get@@entry.type.otf2@@ = conf.lib.OTF2_AttributeList_Get@@entry.type.otf2@@
    c_Get@@entry.type.otf2@@.argtypes = [ ctypes.POINTER(AttributeList), AttributeRef, ctypes.POINTER(@@entry.type.py_argtype@@) ]
    c_Get@@entry.type.otf2@@.restype = ErrorCode
    c_Get@@entry.type.otf2@@.errcheck = HandleErrorCode
    @@entry.type.lower@@Value = @@entry.type.py_ctype@@()
    c_Get@@entry.type.otf2@@(attributeList, attribute, ctypes.byref(@@entry.type.lower@@Value))
    return @@entry.type.lower@@Value.value

@otf2 endfor
def AttributeList_GetString(attributeList, attribute):
    c_GetString = conf.lib.OTF2_AttributeList_GetString
    c_GetString.argtypes = [ ctypes.POINTER(AttributeList), AttributeRef, ctypes.POINTER(StringRef) ]
    c_GetString.restype = ErrorCode
    c_GetString.errcheck = HandleErrorCode
    stringRef = StringRef()
    c_GetString(attributeList, attribute, ctypes.byref(stringRef))
    return stringRef.value

@otf2 for entry in MappingType.entries:
@otf2  with def = entry.def
def AttributeList_Get@@def.ref_name@@Ref(attributeList, attribute):
    c_Get@@def.ref_name@@Ref = conf.lib.OTF2_AttributeList_Get@@def.ref_name@@Ref
    c_Get@@def.ref_name@@Ref.argtypes = [ ctypes.POINTER(AttributeList), AttributeRef, ctypes.POINTER(@@def.self_attr.type.py_argtype@@) ]
    c_Get@@def.ref_name@@Ref.restype = ErrorCode
    c_Get@@def.ref_name@@Ref.errcheck = HandleErrorCode
    @@def.ref_lname@@Ref = @@def.self_attr.type.py_ctype@@()
    c_Get@@def.ref_name@@Ref(attributeList, attribute, ctypes.byref(@@def.ref_lname@@Ref))
    return @@def.ref_lname@@Ref.value

@otf2  endwith
@otf2 endfor
def AttributeList_GetAttributeByIndex(attributeList, index, attribute = AttributeRef(),
                                      attributeType = Type(),attributeValue = AttributeValue()):
    c_GetAttributeByIndex = conf.lib.OTF2_AttributeList_GetAttributeByIndex
    c_GetAttributeByIndex.argtypes = [ ctypes.POINTER(AttributeList), ctypes.c_uint32, ctypes.POINTER(AttributeRef), ctypes.POINTER(Type), ctypes.POINTER(AttributeValue) ]
    c_GetAttributeByIndex.restype = ErrorCode
    c_GetAttributeByIndex.errcheck = HandleErrorCode
    c_GetAttributeByIndex(attributeList, index, ctypes.byref(attribute), ctypes.byref(attributeType), ctypes.byref(attributeValue))
    return (attribute.value, attributeType, attributeValue)

def AttributeList_PopAttribute(attributeList, attribute = AttributeRef(),
                               attributeType = Type(), attributeValue = AttributeValue()):
    c_PopAttribute = conf.lib.OTF2_AttributeList_PopAttribute
    c_PopAttribute.argtypes = [ ctypes.POINTER(AttributeList), ctypes.POINTER(AttributeRef), ctypes.POINTER(Type), ctypes.POINTER(AttributeValue) ]
    c_PopAttribute.restype = ErrorCode
    c_PopAttribute.errcheck = HandleErrorCode
    c_PopAttribute(attributeList, ctypes.byref(attribute), ctypes.byref(attributeType), ctypes.byref(attributeValue))
    return (attribute.value, attributeType, attributeValue)

def AttributeList_GetNumberOfElements(attributeList):
    c_GetNumberOfElements = conf.lib.OTF2_AttributeList_GetNumberOfElements
    c_GetNumberOfElements.argtype = [ ctypes.POINTER(AttributeList) ]
    c_GetNumberOfElements.restype = ctypes.c_uint32
    return c_GetNumberOfElements(attributeList)

__all__ = [
    'AttributeList',
    'AttributeList_New',
    'AttributeList_Delete',
    'AttributeList_AddAttribute',
    'AttributeList_AddString',
    @otf2 for entry in Type.entries if entry.type:
    'AttributeList_Add@@entry.type.otf2@@',
    'AttributeList_Get@@entry.type.otf2@@',
    @otf2 endfor
    'AttributeList_RemoveAttribute',
    'AttributeList_RemoveAllAttributes',
    'AttributeList_TestAttributeByID',
    'AttributeList_GetAttributeByID',
    @otf2 for entry in MappingType.entries:
    @otf2 with def = entry.def
    'AttributeList_Add@@entry.def.ref_name@@Ref',
    'AttributeList_Get@@def.ref_name@@Ref',
    @otf2 endwith
    @otf2 endfor
    'AttributeList_GetString',
    'AttributeList_GetAttributeByIndex',
    'AttributeList_PopAttribute',
    'AttributeList_GetNumberOfElements'
]



#! /usr/bin/env @PYTHON_FOR_GENERATOR@

##
## This file is part of the Score-P software (http://www.score-p.org)
##
## Copyright (c) 2011-2014, 2017, 2020, 2024,
## Technische Universitaet Dresden, Germany
##
## This software may be modified and distributed under the terms of
## a BSD-style license.  See the COPYING file in the package base
## directory for details.
##

from __future__ import print_function

import sys
import os
import re
import textwrap
import traceback

class Otf2Version(object):
    """Object to holds the OTF2 version information."""
    def __init__(self, major, minor, bugfix, suffix, version, interface_current, interface_revision, interface_age):
        self.major = major
        self.minor = minor
        self.bugfix = bugfix
        self.suffix = suffix
        self.version = version
        self.soversion = interface_current - interface_age
        self.interface = '%d:%d:%d' % (interface_current, interface_revision, interface_age,)

    def __str__(self):
        return str(self.version)

prefix = '@prefix@'
exec_prefix = '@exec_prefix@'
datarootdir = '@datarootdir@'
package = '@PACKAGE@'
if prefix[1:-1] == 'prefix':
    try:
        abs_builddir = os.getcwd()
        mypath = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        abs_srcdir = os.path.dirname(mypath)
    except:
        print("Can't determine build and source directory! Can only be called from a build dir.", file=sys.stderr)
        sys.exit(1)

    # add our otf2 module to the system module load path
    sys.path.insert(1, os.path.join(abs_srcdir, 'share'))

    # add our otf2 module from the build dir to the system module load path
    sys.path.insert(1, os.path.join(abs_builddir, 'share', 'otf2'))

else:
    # resolve ${exec_prefix} and ${prefix}
    if datarootdir[:14] == '${exec_prefix}':
        datarootdir = exec_prefix + datarootdir[14:]
    if datarootdir[:9] == '${prefix}':
        datarootdir = prefix + datarootdir[9:]

    # add our otf2 module to the system module load path
    sys.path.insert(1, os.path.join(datarootdir, package, 'python'))

import otf2
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateError, TemplateNotFound, TemplateSyntaxError, UndefinedError

env = Environment(
    loader=FileSystemLoader(['.', '/']),
    line_statement_prefix='@otf2',
    line_comment_prefix='//@',
    block_start_string='{{@',
    block_end_string='@}}',
    variable_start_string='@@',
    variable_end_string='@@',
    comment_start_string='/*@',
    comment_end_string='@*/',
    keep_trailing_newline=True
)

def jinja2filter(env):
    def make_filter(f):
        assert f.__name__[:7] == 'filter_', 'Function should start with filter_'
        env.filters[f.__name__[7:]] = f
        return f
    return make_filter

def jinja2test(env):
    def make_test(f):
        assert f.__name__[:5] == 'test_', 'Function should start with test_'
        env.tests[f.__name__[5:]] = f
        return f
    return make_test

@jinja2filter(env)
def filter_exclude(values, *excludes):
    s = set(excludes)
    return [v for v in values if v not in s]

@jinja2filter(env)
def filter_include(values, *includes):
    s = set(includes)
    return [v for v in values if v in s]

@jinja2test(env)
def test_snapshot_event(e):
    assert isinstance(e, otf2.EventRecord), 'Only event records allowed'
    return hasattr(e, 'snapshot_record') and e.snapshot_record

@jinja2filter(env)
def filter_snapshot_events(events):
    return [e for e in events if test_snapshot_event(e)]

@jinja2test(env)
def test_global_only_def(d):
    assert isinstance(d, otf2.DefRecord), 'Only definitions allowed.'
    return isinstance(d, otf2.GlobalOnlyDefRecord)

@jinja2filter(env)
def filter_global_only_defs(defs):
    return [d for d in defs if test_global_only_def(d)]

@jinja2filter(env)
def filter_not_global_only_defs(defs):
    return [d for d in defs if not test_global_only_def(d)]

@jinja2test(env)
def test_global_def(d):
    assert isinstance(d, otf2.DefRecord), 'Only definitions allowed.'
    return not isinstance(d, otf2.LocalOnlyDefRecord)

@jinja2filter(env)
def filter_global_defs(defs):
    return [d for d in defs if test_global_def(d)]

@jinja2filter(env)
def filter_not_global_defs(defs):
    return [d for d in defs if not test_global_def(d)]

@jinja2test(env)
def test_local_def(d):
    assert isinstance(d, otf2.DefRecord), 'Only definitions allowed.'
    return not isinstance(d, otf2.GlobalOnlyDefRecord)

@jinja2filter(env)
def filter_local_defs(defs):
    return [d for d in defs if test_local_def(d)]

@jinja2filter(env)
def filter_not_local_defs(defs):
    return [d for d in defs if not test_local_def(d)]

@jinja2test(env)
def test_local_only_def(d):
    assert isinstance(d, otf2.DefRecord), 'Only definitions allowed.'
    return isinstance(d, otf2.LocalOnlyDefRecord)

@jinja2filter(env)
def filter_local_only_defs(defs):
    return [d for d in defs if test_local_only_def(d)]

@jinja2filter(env)
def filter_not_local_only_defs(defs):
    return [d for d in defs if not test_local_only_def(d)]

@jinja2test(env)
def test_first_born_def(d):
    assert isinstance(d, otf2.DefRecord), 'Only definitions allowed.'
    return d.siblings[0] == d

@jinja2filter(env)
def filter_first_born_defs(defs):
    return [d for d in defs if test_first_born_def(d)]

@jinja2filter(env)
def filter_not_first_born_defs(defs):
    return [d for d in defs if not test_first_born_def(d)]

@jinja2test(env)
def test_shared_def(d):
    assert isinstance(d, otf2.DefRecord), 'Only definitions allowed.'
    return len(d.siblings) > 1

@jinja2filter(env)
def filter_shared_defs(defs):
    return [d for d in defs if test_shared_def(d)]

@jinja2filter(env)
def filter_not_shared_defs(defs):
    return [d for d in defs if not test_shared_def(d)]

@jinja2test(env)
def test_referable_def(d):
    assert isinstance(d, otf2.DefRecord), 'Only definitions allowed.'
    return d.self_attr is not None

@jinja2filter(env)
def filter_referable_defs(defs):
    return [d for d in defs if test_referable_def(d)]

@jinja2filter(env)
def filter_not_referable_defs(defs):
    return [d for d in defs if not test_referable_def(d)]

@jinja2test(env)
def test_supplement_def(d):
    assert isinstance(d, otf2.DefRecord), 'Only definitions allowed.'
    return isinstance(d, otf2.SuppDefRecord)

@jinja2filter(env)
def filter_supplement_defs(defs):
    return [d for d in defs if test_supplement_def(d)]

@jinja2filter(env)
def filter_not_supplement_defs(defs):
    return [d for d in defs if not test_supplement_def(d)]

@jinja2filter(env)
def filter_pluralize(name):
    return otf2.pluralize(name)

@jinja2test(env)
def test_bitset_enum(e):
    assert isinstance(e, otf2.Enum), 'Only enums allowed.'
    return e.bitset_like

@jinja2test(env)
def test_general_enum(e):
    assert isinstance(e, otf2.Enum), 'Only enums allowed.'
    return isinstance(e, otf2.GeneralEnum)

@jinja2filter(env)
def filter_general_enums(enums):
    return [e for e in enums if test_general_enum(e)]

@jinja2test(env)
def test_def_enum(e):
    assert isinstance(e, otf2.Enum), 'Only enums allowed.'
    return isinstance(e, otf2.DefEnum)

@jinja2filter(env)
def filter_def_enums(enums):
    return [e for e in enums if test_def_enum(e)]

@jinja2test(env)
def test_py_def_enum_entries(e):
    if isinstance(e.value, str):
        if 'OTF2_' in e.value:
                    return False
    return isinstance(e, otf2.EnumEntry)

@jinja2filter(env)
def filter_py_def_enum_entries(entries):
    return [e for e in entries if test_py_def_enum_entries(e)]

@jinja2test(env)
def test_event_enum(e):
    assert isinstance(e, otf2.Enum), 'Only enums allowed.'
    return isinstance(e, otf2.EventEnum)

@jinja2filter(env)
def filter_event_enums(enums):
    return [e for e in enums if test_event_enum(e)]

@jinja2test(env)
def test_enum_with_terminator(e):
    assert isinstance(e, otf2.Enum), 'Only enums allowed.'
    return hasattr(e, 'terminator') and e.terminator is not None

@jinja2test(env)
def test_array_attr(a):
    assert isinstance(a, otf2.Attribute), 'Only attributes allowed.'
    return isinstance(a, otf2.ArrayAttribute)

@jinja2test(env)
def test_def_ref_attr(a):
    assert isinstance(a, otf2.Attribute), 'Only attributes allowed.'
    return isinstance(a, otf2.DefRefAttribute)

@jinja2test(env)
def test_enum_attr(a):
    assert isinstance(a, otf2.Attribute), 'Only attributes allowed.'
    return isinstance(a, otf2.EnumAttribute)

@jinja2test(env)
def test_attic_attr(a):
    assert isinstance(a, otf2.Attribute), 'Only attributes allowed.'
    return hasattr(a, 'attic') and a.attic is not None

@jinja2test(env)
def test_singleton_event(e):
    assert isinstance(e, otf2.EventRecord), 'Only events allowed.'
    return hasattr(e, 'singleton') and e.singleton

@jinja2test(env)
def test_derived_snap(e):
    assert isinstance(e, otf2.SnapshotRecord), 'Only snapshots allowed.'
    return hasattr(e, 'origin')

@jinja2test(env)
def test_standard_type(t):
    return t in ['int8_t', 'uint8_t', 'int16_t', 'uint16_t', 'int32_t', 'uint32_t', 'int64_t', 'uint64_t', 'double', 'const char*']

tmpl_data = {
    'doxygenparams': otf2.doxygenparams,
    'version': Otf2Version(
        otf2.__version_info__[0],
        otf2.__version_info__[1],
        otf2.__version_info__[2],
        otf2.__version_info__[3],
        otf2.__version__,
        otf2.__interface__[0],
        otf2.__interface__[1],
        otf2.__interface__[2]),
    'types': otf2.types,
    'enums': otf2.enums,
    'defs': otf2.defs,
    'events': otf2.events,
    'snaps': otf2.snaps,
    'attic': {
        'enums': otf2.attic_enums
    }
}
for e in otf2.types:
    tmpl_data[str(e)] = e
for e in otf2.enums:
    tmpl_data[str(e)] = e
    tmpl_data[str(e)+'Enum'] = e
for e in otf2.defs:
    if str(e) not in tmpl_data:
        tmpl_data[str(e)] = e
    tmpl_data[str(e)+'Def'] = e
for e in otf2.events:
    if str(e) not in tmpl_data:
        tmpl_data[str(e)] = e
    tmpl_data[str(e)+'Event'] = e
for e in otf2.snaps:
    if str(e) not in tmpl_data:
        tmpl_data[str(e)] = e
    tmpl_data[str(e)+'Snap'] = e

if '--help' in sys.argv \
        or '-h'  in sys.argv:
    print('Usage: otf2-template TEMPLATE-FILE [OUTPUT-FILE]')
    print('Template driver for the OTF2 library.')
    print()
    print('Report bugs to <%s>' % ('@PACKAGE_BUGREPORT@',))
    sys.exit(0)
elif '--version' in sys.argv \
        or '-V'  in sys.argv:
    print('otf2-template: version %s' % (otf2.__version__,))
    sys.exit(0)

if len(sys.argv) not in (2, 3):
    print('otf2-template: unsupported number of arguments', file=sys.stderr)
    print('Try \'otf2-template --help\' for more information.', file=sys.stderr)
    sys.exit(1)

tmplfile = sys.argv[1]
tmplname = os.path.abspath(tmplfile)
try:
    tmpl = env.get_template(tmplname)
except TemplateNotFound as e:
    print('otf2-template: can\'t load template \'%s\': %s' % (tmplfile, e), file=sys.stderr)
    sys.exit(1)
except TemplateSyntaxError as e:
    print('%s:%d: error: %s' % (tmplfile, e.lineno, e.message), file=sys.stderr)
    sys.exit(1)
except Exception as e:
    print('otf2-template: can\'t load template \'%s\': %s' % (tmplfile, e), file=sys.stderr)
    sys.exit(1)

if len(sys.argv) == 3:
    try:
        out = open(sys.argv[2], 'w')
    except Exception as e:
        print('otf2-template: can\'t open output file \'%s\': %s' % (sys.argv[2], e,), file=sys.stderr)
        sys.exit(1)
else:
    out = sys.stdout

try:
    tmpl.stream(tmpl_data).dump(out)
except TemplateSyntaxError as e:
    print('%s:%d: %s' % (tmplfile, e.lineno, e.message), file=sys.stderr)
    sys.exit(1)
except UndefinedError as e:
    for filename, lineno, func, text in traceback.extract_tb(sys.exc_info()[2]):
        if filename == tmplname:
            print('%s: In %s:' % (tmplfile, func), file=sys.stderr)
            print('%s:%d: error: %s' % (tmplfile, lineno, e.message), file=sys.stderr)
            sys.exit(1)
    print('%s: %s' % (tmplfile, e.message), file=sys.stderr)
    sys.exit(1)
except TemplateError as e:
    print('%s: %s' % (tmplfile, e.message), file=sys.stderr)
    sys.exit(1)
finally:
    if out != sys.stdout:
        out.close()




""" Generated by OTF2 Template Engine """

import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import *
from .Definitions import *
from .Events import *


class AttributeValue(ctypes.Union):
    """
    ctypes does not support unions as (non-pointer) arguments
    _arg_value / _arg_type is a crude and evil hack to guess
    the right argument type. This type must be correct in terms of
    size (maximum of the union) and
    calling convention (architecture specific, but for AMD64 int usually wins).
    Whenever used as a direct function argument (also applies to callbacks)
    this type should be used instead of the union. Otherwise you get undefined
    behavior.
    """
    _arg_type = ctypes.c_uint64
    _fields_ = [
@otf2 for entry in Type.entries if entry.type:
        ("@@entry.type.lower@@", @@entry.type.py_ctype@@),
@otf2 endfor
@otf2 for entry in MappingType.entries:
        ("@@entry.def.ref_lname@@Ref", @@entry.def.self_attr.type.py_ctype@@),
@otf2 endfor
        ("_arg_value", ctypes.c_uint64),
    ]

@otf2 for enum in enums:
def AttributeValue_Set@@enum.name@@(enumValue, type_=Type(), value=AttributeValue()):
    c_Set@@enum.name@@ = conf.lib.OTF2_AttributeValue_Set@@enum.name@@
    c_Set@@enum.name@@.argtypes = [ @@enum.type.py_argtype@@, ctypes.POINTER(Type), ctypes.POINTER(AttributeValue) ]
    c_Set@@enum.name@@.restype = ErrorCode
    c_Set@@enum.name@@.errcheck = HandleErrorCode
    c_Set@@enum.name@@(enumValue, ctypes.byref(type_), ctypes.byref(value))
    return (type_, value)

@otf2 endfor
@otf2 for enum in enums:
def AttributeValue_Get@@enum.name@@(type, enum_value=@@enum.type.py_ctype@@()):
    c_Get@@enum.name@@ = conf.lib.OTF2_AttributeValue_Get@@enum.name@@
    c_Get@@enum.name@@.argtypes = [ Type, AttributeValue._arg_type, ctypes.POINTER(@@enum.type.py_ctype@@) ]
    c_Get@@enum.name@@.restype = ErrorCode
    c_Get@@enum.name@@.errcheck = HandleErrorCode
    c_Get@@enum.name@@(type, value._arg_value, ctypes.byref(enum_value))
    return enum_value

@otf2 endfor
__all__ = [
    'AttributeValue',
    @otf2 for enum in enums:
    'AttributeValue_Set@@enum.name@@',
    'AttributeValue_Get@@enum.name@@',
    @otf2 endfor
]



class Error(Exception):
    """ Base class for OTF2 high level exceptions. """
    def __str__(self):
        return self.msg


class TraceReaderError(Error):
    def __init__(self, msg):
        self.msg = "Semantic error in the input trace: " + msg


class TraceWriterError(Error):
    def __init__(self, msg):
        self.msg = "Semantic error trying to write the trace: " + msg




# Generated by OTF2 Template Engine

import _otf2


class _EnumWrapper(object):
    def __init__(self, prefix, enum_type=None):
        # This is a bit ugly, but IMHO at least less ugly than doing it in __getattr__
        self.__dict__['type'] = enum_type
        for (name, candidate) in _otf2.__dict__.items():
            if (enum_type is None or isinstance(candidate, enum_type)) and name.startswith(prefix):
                short_name = name[len(prefix):]
                self.__dict__[short_name] = candidate

    def __setattr__(self, name, value):
        raise AttributeError("Change of values is not allowed.")


# GeneralDefinitions
Undefined = _EnumWrapper("UNDEFINED_")

__all__ = ["Undefined", "FileMode", "FlushType", "Compression"]

FileMode = _otf2.FileMode
FlushType = _otf2.FlushType
Compression = _otf2.Compression
@otf2 for enum in enums:

@@enum.name@@ = _otf2.@@enum.name@@
__all__.append("@@enum.name@@")
@otf2 endfor



common_template_deps = [
    'share/otf2/otf2.types',
    'share/otf2/otf2.enums',
    'share/otf2/otf2.defs',
    'share/otf2/otf2.events',
    'share/otf2/otf2.snaps'
]

templates = [
    [ 'include/otf2/OTF2_GeneralDefinitions.h',
        'OTF2_GeneralDefinitions.tmpl.h',
        'build-config/VERSION' ],
    [ 'include/otf2/OTF2_Definitions.h',
        'OTF2_Definitions.tmpl.h' ],
    [ 'src/otf2_file_types.h',
        'otf2_file_types.tmpl.h' ],
    [ 'include/otf2/OTF2_Events.h',
        'OTF2_Events.tmpl.h' ],
    [ 'src/otf2_events_int.h',
        'otf2_events_int.tmpl.h' ],
    [ 'include/otf2/OTF2_EvtReaderCallbacks.h',
        'OTF2_EvtReaderCallbacks.tmpl.h' ],
    [ 'src/OTF2_EvtReader_inc.c',
        'OTF2_EvtReader_inc.tmpl.c' ],
    [ 'include/otf2/OTF2_EvtWriter.h',
        'OTF2_EvtWriter.tmpl.h' ],
    [ 'src/OTF2_EvtWriter_inc.c',
        'OTF2_EvtWriter_inc.tmpl.c' ],
    [ 'include/otf2/OTF2_GlobalEvtReaderCallbacks.h',
        'OTF2_GlobalEvtReaderCallbacks.tmpl.h' ],
    [ 'include/otf2/OTF2_GlobalSnapReaderCallbacks.h',
        'OTF2_GlobalSnapReaderCallbacks.tmpl.h' ],
    [ 'src/OTF2_GlobalEvtReader_inc.c',
        'OTF2_GlobalEvtReader_inc.tmpl.c' ],
    [ 'src/OTF2_GlobalSnapReader_inc.c',
        'OTF2_GlobalSnapReader_inc.tmpl.c' ],
    [ 'src/otf2_local_definitions.h',
        'otf2_local_definitions.tmpl.h' ],
    [ 'include/otf2/OTF2_DefReaderCallbacks.h',
        'OTF2_DefReaderCallbacks.tmpl.h' ],
    [ 'src/OTF2_DefReader_inc.c',
        'OTF2_DefReader_inc.tmpl.c' ],
    [ 'include/otf2/OTF2_DefWriter.h',
        'OTF2_DefWriter.tmpl.h' ],
    [ 'src/OTF2_DefWriter_inc.c',
        'OTF2_DefWriter_inc.tmpl.c' ],
    [ 'src/otf2_global_definitions.h',
        'otf2_global_definitions.tmpl.h' ],
    [ 'include/otf2/OTF2_GlobalDefReaderCallbacks.h',
        'OTF2_GlobalDefReaderCallbacks.tmpl.h' ],
    [ 'src/OTF2_GlobalDefReader_inc.c',
        'OTF2_GlobalDefReader_inc.tmpl.c' ],
    [ 'include/otf2/OTF2_GlobalDefWriter.h',
        'OTF2_GlobalDefWriter.tmpl.h' ],
    [ 'src/OTF2_GlobalDefWriter_inc.c',
        'OTF2_GlobalDefWriter_inc.tmpl.c' ],
    [ 'src/otf2_snaps.h',
        'otf2_snaps.tmpl.h' ],
    [ 'include/otf2/OTF2_SnapReaderCallbacks.h',
        'OTF2_SnapReaderCallbacks.tmpl.h' ],
    [ 'src/OTF2_SnapReader_inc.c',
        'OTF2_SnapReader_inc.tmpl.c' ],
    [ 'include/otf2/OTF2_SnapWriter.h',
        'OTF2_SnapWriter.tmpl.h' ],
    [ 'src/OTF2_SnapWriter_inc.c',
        'OTF2_SnapWriter_inc.tmpl.c' ],
    [ 'src/tools/otf2_print/otf2_print_inc.c',
        'otf2_print_inc.tmpl.c' ],
    [ 'src/tools/otf2_print/otf2_print_types.h',
        'otf2_print_types.tmpl.h' ],
    [ 'src/tools/otf2_snapshots/otf2_snapshots_events.hpp',
        'otf2_snapshots_events.tmpl.hpp' ],
    [ 'src/tools/otf2_snapshots/otf2_snapshots_events_inc.cpp',
        'otf2_snapshots_events_inc.tmpl.cpp' ],
    [ 'src/tools/otf2_snapshots/otf2_snapshots_callbacks_inc.cpp',
        'otf2_snapshots_callbacks_inc.tmpl.cpp' ],
    [ 'src/otf2_attribute_value_inc.c',
        'otf2_attribute_value_inc.tmpl.c' ],
    [ 'include/otf2/OTF2_AttributeValue.h',
        'OTF2_AttributeValue.tmpl.h' ],
    [ 'src/otf2_attribute_list_inc.c',
        'otf2_attribute_list_inc.tmpl.c' ],
    [ 'include/otf2/OTF2_AttributeList.h',
        'OTF2_AttributeList.tmpl.h' ],
    [ 'src/otf2_attic_types.h',
        'otf2_attic_types.tmpl.h' ],
    [ 'doc/doxygen-user/definitions.dox',
        'definitions.tmpl.dox' ],
    [ 'doc/doxygen-user/events.dox',
        'events.tmpl.dox' ],
    [ 'doc/doxygen-user/snaps.dox',
        'snaps.tmpl.dox' ],
    [ 'templates/example.cc',
        'example.tmpl.cc' ],
    [ 'include/otf2/OTF2_EventSizeEstimator.h',
        'OTF2_EventSizeEstimator.tmpl.h' ],
    [ 'src/OTF2_EventSizeEstimator.c',
        'OTF2_EventSizeEstimator.tmpl.c' ],
    [ 'src/tools/otf2_estimator/otf2_estimator_inc.cpp',
        'otf2_estimator_inc.tmpl.cpp' ],
    # python bindings
    [ 'src/python/_otf2/Definitions.py',
            '_otf2.Definitions.tmpl.py' ],
    [ 'src/python/_otf2/Events.py',
            '_otf2.Events.tmpl.py' ],
    [ 'src/python/_otf2/GeneralDefinitions.py',
            '_otf2.GeneralDefinitions.tmpl.py', ],
    [ 'src/python/_otf2/GlobalDefReaderCallbacks.py',
            '_otf2.GlobalDefReaderCallbacks.tmpl.py' ],
    [ 'src/python/_otf2/DefReaderCallbacks.py',
            '_otf2.DefReaderCallbacks.tmpl.py' ],
    [ 'src/python/_otf2/EvtReaderCallbacks.py',
            '_otf2.EvtReaderCallbacks.tmpl.py' ],
    [ 'src/python/_otf2/GlobalEvtReaderCallbacks.py',
            '_otf2.GlobalEvtReaderCallbacks.tmpl.py' ],
    [ 'src/python/_otf2/GlobalSnapReaderCallbacks.py',
            '_otf2.GlobalSnapReaderCallbacks.tmpl.py' ],
    [ 'src/python/_otf2/SnapReaderCallbacks.py',
            '_otf2.SnapReaderCallbacks.tmpl.py' ],
    [ 'src/python/_otf2/EvtWriter.py',
            '_otf2.EvtWriter.tmpl.py' ],
    [ 'src/python/_otf2/DefWriter.py',
            '_otf2.DefWriter.tmpl.py' ],
    [ 'src/python/_otf2/GlobalDefWriter.py',
            '_otf2.GlobalDefWriter.tmpl.py' ],
    [ 'src/python/_otf2/SnapWriter.py',
            '_otf2.SnapWriter.tmpl.py' ],
    [ 'src/python/_otf2/AttributeList.py',
            '_otf2.AttributeList.tmpl.py' ],
    [ 'src/python/_otf2/EventSizeEstimator.py',
            '_otf2.EventSizeEstimator.tmpl.py' ],
    [ 'src/python/_otf2/AttributeValue.py',
            '_otf2.AttributeValue.tmpl.py' ],
    [ 'src/python/otf2/events.py',
            'otf2.events.tmpl.py' ],
    [ 'src/python/otf2/event_reader.py',
            'otf2.event_reader.tmpl.py' ],
    [ 'src/python/otf2/event_writer.py',
            'otf2.event_writer.tmpl.py' ],
    [ 'src/python/otf2/registry.py',
            'otf2.registry.tmpl.py' ],
    [ 'src/python/otf2/definition_writer.py',
            'otf2.definition_writer.tmpl.py' ],
    [ 'src/python/otf2/enums.py',
            'otf2.enums.tmpl.py' ],
    [ 'src/python/otf2/attribute_value.py',
            'otf2.attribute_value.tmpl.py' ],
    [ 'contrib/CMakeLists.txt',
        'contrib-CMakeLists.tmpl.txt',
        'build-config/VERSION' ],
    [ 'contrib/config-backend.h',
        'contrib-config-backend.tmpl.h',
        'build-config/VERSION' ],
]




''' Generated by OTF2 Template Engine '''

import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import *
from .AttributeList import AttributeList
from .Definitions import *
from .Events import *


class SnapWriter(ctypes.Structure):
    pass

def SnapWriter_GetLocationID(writer, locationID = LocationRef()):
    c_GetLocationID = conf.lib.OTF2_SnapWriter_GetLocationID
    c_GetLocationID.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(LocationRef) ]
    c_GetLocationID.restype = ErrorCode
    c_GetLocationID.errcheck = HandleErrorCode
    c_GetLocationID(writer, ctypes.byref(locationID))
    return locationID.value

@otf2 for snap in snaps:
@otf2  if snap == MetricSnap:
def SnapWriter_Metric(writerHandle, attributeList, snapTime, origEventTime, metric, typeIDs, metricValues):
    numberOfMetrics = len(metricValues)
    assert(numberOfMetrics == len(typeIDs))

    TypeArrayType = Type * numberOfMetrics
    type_ids_list = TypeArrayType()
    type_ids_list[:] = typeIDs

    MetricValueArrayType = MetricValue * numberOfMetrics
    metric_values_list = MetricValueArrayType()
    metric_values_list[:] = metricValues

    c_Metric = conf.lib.OTF2_SnapWriter_Metric
    c_Metric.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp , TimeStamp, MetricRef, ctypes.c_uint8, TypeArrayType, MetricValueArrayType ]
    c_Metric.restype = ErrorCode
    c_Metric.errcheck = HandleErrorCode
    c_Metric(writerHandle, attributeList, snapTime , origEventTime, metric, numberOfMetrics, type_ids_list, metric_values_list)

@otf2  else:
def SnapWriter_@@snap.name@@(writerHandle, attributeList, snapTime@@snap.py_funcargs()@@):
    c_@@snap.name@@ = conf.lib.OTF2_SnapWriter_@@snap.name@@
    c_@@snap.name@@.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp@@snap.py_funcargtypes(argtypes=False)@@ ]
    c_@@snap.name@@.restype = ErrorCode
    c_@@snap.name@@.errcheck = HandleErrorCode
    c_@@snap.name@@(writerHandle, attributeList, snapTime@@snap.py_callargs(pass_by_ref=True)@@)

@otf2  endif
@otf2 endfor
__all__ = [
    'SnapWriter',
    'SnapWriter_GetLocationID',
    @otf2 for snap in snaps:
    'SnapWriter_@@snap.name@@',
    @otf2 endfor
]



import unittest


class BasicImportTest(unittest.TestCase):
    @staticmethod
    def test_import_ll():
        import _otf2

    @staticmethod
    def test_import_hl():
        import otf2


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python

import sys
import os
import re
import textwrap

# we are called from the src directory directly, use path relative to
# our location, we sit in templates/
mypath = os.path.dirname(os.path.abspath(__file__))

from jinja2 import Environment, FileSystemLoader

def _execfile(filepath, globals=globals(), locals=None):
    if globals is None:
        globals = {}
    globals.update({
        "__file__": filepath,
        "__name__": "__main__",
    })
    with open(filepath, 'rb') as file:
        exec(compile(file.read(), filepath, 'exec'), globals, locals)

env = Environment(
    loader=FileSystemLoader(mypath),
    line_statement_prefix='#tmpl',
    line_comment_prefix='###',
    block_start_string='{#',
    block_end_string='#}',
    variable_start_string='##',
    variable_end_string='##',
    comment_start_string='/*#',
    comment_end_string='#*/',
    keep_trailing_newline=True
)

def jinja2filter(env):
    def make_filter(f):
        assert f.__name__[:7] == 'filter_', 'Function should start with filter_'
        env.filters[f.__name__[7:]] = f
        return f
    return make_filter

@jinja2filter(env)
def filter_prefix_path(values, prefix=' '):
    if len(values) == 0:
        return ''
    return prefix + prefix.join(values)

_execfile(os.path.join(mypath, 'templates.py'))

tmpl = env.get_template(os.path.basename(sys.argv[1]))
out = open(sys.argv[2], 'w')

tmpl.stream({
    'common_template_deps': common_template_deps,
    'templates':            templates
}).dump(out)
out.close()



# Generated by OTF2 Template Engine

import re
from numbers import Number, Integral

from .attribute_list import AttributeList
from .enums import Base, ParameterType


def _replace_doxygen_commands(doc_string):
    if doc_string is not None:
        p = re.compile(r'@eref\{([^}]*)\}')
        doc_string = p.sub(r':py:class:`\1`', doc_string)
        doc_string = doc_string.replace('@b', '')
    return doc_string


class _EventMeta(type):
    def __new__(mcls, name, bases, namespace):
        clazz = type.__new__(mcls, name, bases, namespace)
        old_doc = namespace.get("__doc__")
        clazz.__doc__ = _replace_doxygen_commands(old_doc)
        return clazz


class _Event(metaclass=_EventMeta):
    def __init__(self, time, attributes=None):
        self.time = time
        if attributes is not None:
            if not isinstance(attributes, dict):
                raise TypeError("Unexpected type for argument: attributes")
        self.attributes = attributes

    def __str__(self):
        return str(self.__dict__)


def _get_string_ref(writer, value):
    # The writer argument is the event writer, but for get_ref its the definition
    # writer. But we do not want, that this string definition will be written now
    return writer._archive.definitions.strings.get_ref(value, writer=None)
@otf2 for event in events:
@otf2  if event == Metric:

class Metric(_Event):
    """
        This event represents a measurement point.
    """

    _fields = ("time", "int", "the timestamp"), \
              ("metric", "MetricRef", "the :py:class:`otf2.definitions.MetricClass` or \
               :py:class:`otf2.definitions.MetricInstance`"), \
              ("values", "array", "an array of values, one for each :py:class:`otf2.definitions.MetricMember`"), \
              ("attributes", "dict", "a dict with an :py:class:`otf2.definitions.Attribute` \
               referencing a suitable value.")

    def __init__(self, time, metric, values, attributes=None):
        super(Metric, self).__init__(time, attributes)
        self.metric = metric
        if isinstance(values, Number):
            self.values = [values]
        else:
            self.values = values

    @property
    def value(self):
        if len(self.values) != 1:
            raise AttributeError(
                "Trying to access singular metric value on metric event with {} values".format(
                    len(self.values)))
        return self.values[0]

    @value.setter
    def value(self, new_value):
        if len(self.values) != 1:
            raise AttributeError(
                "Trying to access singular metric value on metric event with {} values".format(
                    len(self.values)))
        self.values[0] = new_value

    @property
    def member(self):
        if len(self.values) != 1:
            raise AttributeError(
                "Trying to access singular metric value on metric event with {} values".format(
                    len(self.values)))
        return self.metric.members[0]

    @property
    def scaled_value(self):
        if self.member.base == Base.DECIMAL:
            base = 10
        else:
            base = 2

        return self.value * pow(base, self.member.exponent)

    @property
    def scaled_values(self):
        def get_base(base):
            if base == Base.DECIMAL:
                return 10
            return 2

        return [value * pow(get_base(member.base), member.exponent) for value, member in
                zip(self.values, self.metric.members)]

    @classmethod
    def _construct(cls, registry, time, metric_ref, type_ids, union_values):
        # Type ids is redundant with the ids from the metric itself, we don't check the consistency
        # for performance reaons
        metric = registry.metrics[metric_ref]
        values = metric._convert_unions(union_values)
        return cls(time, metric, values)

    def _write(self, writer):
        with AttributeList(writer._archive.definitions, self.attributes) as al:
            writer._metric(al._handle, self.time, self.metric._ref, self.metric._type_ids,
                           self.metric._convert_values(self.values))
@otf2  elif event == ParameterString:
class ParameterString(_Event):
    """A ParameterString record marks that in the current region, the specified
string parameter has the specified value.

    """
    _fields = ("time", "int", "the timestamp"), \
              ("parameter", "ParameterRef", ""), \
              ("string", "StringRef", ""), \
              ("attributes", "dict", "a dict with an :py:class:`otf2.definitions.Attribute` \
               referencing a suitable value.")

    def __init__(self, time, parameter, string, attributes=None):
        super(ParameterString, self).__init__(time, attributes)
        if parameter.parameter_type != ParameterType.STRING:
            raise TypeError("Expected parameter of type STRING, but got {}".format(parameter.parameter_type))
        self.parameter = parameter
        self.string = string

    @classmethod
    def _construct(cls, registry, time, parameter, string):
        return cls(time, registry.parameters[parameter], registry.strings[string])

    def _write(self, writer):
        with AttributeList(writer._archive.definitions, self.attributes) as al:
            writer._parameter_string(al._handle, self.time, self.parameter._ref, _get_string_ref(writer, self.string))
@otf2  elif event == ProgramBegin:

class ProgramBegin(_Event):
    """The ProgramBegin record marks the begin of the program.

This event is restricted to happen at most once on any @eref{Location}
in a @eref{LocationGroup} that is of type
@eref{OTF2_LOCATION_GROUP_TYPE_PROCESS}.

If there is a ProgramBegin record, a corresponding
@eref{ProgramEnd} record on any @eref{Location} in the same
@eref{LocationGroup} is mandatory and vice versa.

None of the timestamps recorded within the same
@eref{LocationGroup} must be smaller than ProgramBegin's timestamp.

    """
    _fields = ("time", "int", "the timestamp"), \
              ("programName", "StringRef", ""), \
              ("programArguments", "array", "an array of program arguments"), \
              ("attributes", "dict", "a dict with an :py:class:`otf2.definitions.Attribute` \
               referencing a suitable value.")

    def __init__(self, time, program_name, program_arguments, attributes=None):
        super(ProgramBegin, self).__init__(time, attributes)
        self.program_name = program_name
        self.program_arguments = program_arguments

    @classmethod
    def _construct(cls, registry, time, program_name_ref, program_argument_refs):
        return cls(time, registry.strings[program_name_ref], [registry.strings[ref] for ref in program_argument_refs])

    def _write(self, writer):
        with AttributeList(writer._archive.definitions, self.attributes) as al:
            writer._program_begin(al._handle, self.time,
                                  _get_string_ref(writer, self.program_name),
                                  [_get_string_ref(writer, argument) for argument in self.program_arguments])
@otf2  else:

class @@event.name@@(_Event):
    """@@event.doc@@

    """
    _fields = ("time", "int", "the timestamp"), \
              @otf2 for attr in event.attributes:
              ("@@attr.name@@", "@@attr.type.py_ctype@@", ""), \
              @otf2 endfor
              ("attributes", "dict", "a dict with an :py:class:`otf2.definitions.Attribute` \
               referencing a suitable value.")

    def __init__(self, time@@event.py_funcargs(case='lower')@@, attributes=None):
        super(@@event.name@@, self).__init__(time, attributes)
        @otf2 for attr in event.attributes:
        @otf2  if attr.py_parameter_check:
        if @@attr.py_parameter_check[0]@@:
            raise @@attr.py_parameter_check[1]@@
        @otf2  endif
        self.@@attr.lower@@ = @@attr.lower@@
        @otf2 endfor

    @classmethod
    def _construct(cls, registry, time@@event.py_funcargs(case='lower')@@):
        return cls(time@@event.formatargs('{lower}',
                                          format_refs='registry.{def_ref.ref_lower}s[{lower}]')@@)

    def _write(self, writer):
        with AttributeList(writer._archive.definitions, self.attributes) as al:
            writer._@@event.lower@@(al._handle, self.time@@event.formatargs('self.{lower}', format_refs='self.{lower}._ref')@@)
@otf2  endif
@otf2 endfor



"""
The bindings are very close to the original OTF2 API (See
https://silc.zih.tu-dresden.de/otf2-current/html/index.html), with the following exceptions:

0. `OTF2_` is omitted from all names. All symbols are directly in the package.
1. Return codes are checked internally and converted to exceptions of type :py:class:`Error`.
   Error codes are never returned by API functions.
2. Array input parameters that are length / pointer in C, are instead a single sequence
   parameter.
3. Output parameters that are handled by pointers, are instead provided as the return value
   of the function.
   In case of multiple returns outputs, a tuple is returned.
   Some functions also allow to hand in the reference parameters, but that does not work
   for basic python types.

Enums
-----

Enums are wrapped types, don't treat them as integers. Just assign and compare for (in)equality.
Use == for comparison, not is.

Callbacks
---------

Callbacks are normal python functions with appropriate arguments similar to those of normal
functions. The caller must ensure that these functions are not garbage collected for the time
that they might be used by OTF2!

.. warning:: DO NOT pass lambdas or bound methods that go out of scope!
    There is no exception here that will save you from attack of the nasal demons.

Strings
-------

`_otf2` uses the native python str type. Internally OTF2 is defined to use
UTF-8 strings. The conversion happens transparently, a user shall hand `str`
objects to the API and will always receive `str` objects by the API.

ctypes
------

`_otf2` uses ctypes. Some of the types of the module are aliases to ctypes-types, but you never
need to explicitly use ctypes.
"""
__copyright__ = "Copyright (c) 2015-2023, Technische Universitaet Dresden, Germany"
__license__ = "BSD"

from .Archive import *
from .AttributeList import *
from .AttributeValue import *
from .Callbacks import *
from .Definitions import *
from .DefReader import *
from .DefReaderCallbacks import *
from .DefWriter import *
from .ErrorCodes import *
from .Events import *
from .EventSizeEstimator import *
from .EvtReader import *
from .EvtReaderCallbacks import *
from .EvtWriter import *
from .GeneralDefinitions import *
from .GlobalDefReader import *
from .GlobalDefReaderCallbacks import *
from .GlobalDefWriter import *
from .GlobalEvtReader import *
from .GlobalEvtReaderCallbacks import *
from .GlobalSnapReader import *
from .GlobalSnapReaderCallbacks import *
from .IdMap import *
from .Marker import *
from .MarkerReader import *
from .MarkerReaderCallbacks import *
from .MarkerWriter import *
from .Reader import *
from .SnapReader import *
from .SnapReaderCallbacks import *
from .SnapWriter import *
from .Thumbnail import *

# From GeneralDefinitions
__version__ = VERSION
